/// @file        TchContr_Fusion.c
/// @copyright     Copyright (c) 2012 Toradex AG \n
///             All information contained herein is, and remains the property of Toradex Ag and its suppliers, if any. The intellectual\n
///             and technical concepts contained herein are proprietary to Toradex AG and its suppliers and may be covered by Swiss and\n
///             foreign patents, patents in process, and are protected by trade secret or copyright law. Dissemination of this information\n
///             or reproduction of this material is strictly forbidden unless prior written permission is obtained from Toradex AG.\n
/// @author      $Author$
/// @version     $Rev$
/// @date        $Date$
/// @brief       Fuction library for the touch controller Fusion
/// @target        Colibri PXA270, PX3xx, T20 WinCE5, WinCE6, WinEC7
/// @caveats    This file is used in the driver and in the driver utilities.\n
///             DRVUTIL markes the use in the utilities (mostly additional messages are printed.)\n
///             DRVUTIL is defined in the project properties ->C/C11++ ->Preprocessor Definition.

#include <windows.h>
#include <winioctl.h>
#include <winuser.h>

#include "StdAfx.h"
#include "TchContr_Fusion.h"
#include "unfd_multitchdrv.h"
#include "IntLib.h"
#include "i2cLib.h"
#include "GpioLib.h"

#if (_WIN32_WCE == 0x600)
/*
 * Touch Input defines and functions
 */

/*
 * Touch input handle
 */
DECLARE_HANDLE(HTOUCHINPUT);

typedef struct tagTOUCHINPUT {
    LONG x;
    LONG y;
    HANDLE hSource;
    DWORD dwID;
    DWORD dwFlags;
    DWORD dwMask;
    DWORD dwTime;
    ULONG_PTR dwExtraInfo;
    DWORD cxContact;
    DWORD cyContact;
} TOUCHINPUT, *PTOUCHINPUT;
typedef TOUCHINPUT const * PCTOUCHINPUT;


/*
 * Conversion of touch input coordinates to pixels
 */
#define TOUCH_COORD_TO_PIXEL(l)         ((l) / 100)


/*
 * Touch input flag values (TOUCHINPUT.dwFlags)
 */

// Specifies that movement occurred. Cannot be combined with TOUCHEVENTF_DOWN.
#define TOUCHEVENTF_MOVE            0x0001

// Specifies that the corresponding touch point was established through a 
// new contact. Cannot be combined with TOUCHEVENTF_MOVE or TOUCHEVENTF_UP.
#define TOUCHEVENTF_DOWN            0x0002

// Specifies that a touch point was removed.
#define TOUCHEVENTF_UP              0x0004

// Specifies that a touch point is in range. Must be set with all 
// TOUCHEVENTF_DOWN and TOUCHEVENTF_MOVE events.
#define TOUCHEVENTF_INRANGE         0x0008

// Specifies that a touch point is designated as primary. Reserved for use
// by the touch proxy.
#define TOUCHEVENTF_PRIMARY         0x0010

// Specifies that this input should not be coalesced in the input stack.
#define TOUCHEVENTF_NOCOALESCE      0x0020

// Specifies that the touch point is associated with a pen contact. 
#define TOUCHEVENTF_PEN             0x0040

// Specifies that the touch event came from the user's palm
#define TOUCHEVENTF_PALM            0x0080

// Specifies to the input system that it is not necessary to further 
// calibrate the x and y coordinates of the sample.
#define TOUCHEVENTF_CALIBRATED      0x1000

// Specifies that the touch point is associated with a symmetric Touch 
// H/W controller like the capacitive technology where the driver cannot 
// accurately pair the X's and the Y's recorded to provide the actual X,Y 
// touch contact points. 
#define TOUCHEVENTF_SYMMETRIC       0x2000

/*
 * Touch input mask values (TOUCHINPUT.dwMask)
 */
#define TOUCHINPUTMASKF_TIMEFROMSYSTEM  0x0001  // the dwTime field contains a system generated value
#define TOUCHINPUTMASKF_EXTRAINFO       0x0002  // the dwExtraInfo field is valid
#define TOUCHINPUTMASKF_CONTACTAREA     0x0004  // the cxContact and cyContact fields are valid

#endif

//==== Debug Zones ================================================================================
// simple handling: debug zones are forced with TREU or FALSE and need a recompile to be activated
// for sophisticated handling: use macro DEBUGZONE(), DBGPARAM dpCurSettings etc.
// (see MSDN e.g. Win CE 5.0, Debug Zones)
#undef ZONE_ERROR
#undef ZONE_WARN
#undef ZONE_FUNCTION
#undef ZONE_INFO
#undef ZONE_TIPSTATE

#define ZONE_ERROR        TRUE
#define ZONE_WARN         TRUE
#define ZONE_FUNCTION     FALSE
#define ZONE_INFO         FALSE
#define ZONE_TIPSTATE     FALSE


//==== privat preprocessor definitions ============================================================
#ifdef TRIGGER_ON
#define dwTRIGGER_PIN                    106                                // define the pin for trigger
#endif

//-- I2C Communication ----------------------------------------------------------------------------
#define I2C_ADR                            0x10                            // I2C address of the Fusion chip

#define INT_PIN                            133                                // default pin for the "INT Line"
#define INT_SIG_INV                        0                                // 0 => INT signal is as described in the data sheet of the touch controller
                                                                        // >0 => INT signal is inverted
#define RESET_PIN                        127                                // default pin for the "RESET Line"
#define RESET_POSTDLY                    300                                // delay after the "RESET Line" is released
#define RESET_LIN_INV                    0                                // 0 => "RESET Line" is as described in the data sheet of the touch controller
                                                                        // >0 => "RESET line" is inverted

#define I2C_100KB                        1                                // 100kb I2C speed
#define I2C_400KB                        2                                // 400kb I2C speed
#define HWADAPTPRIO                        30  // THREAD_PRIORITY_ABOVE_NORMAL    // task priority hardware adapter

#define    COMTST_INT                        5000                             // timeout wait for comunication test


//-- "CTP Register Copy" --------------------------------------------------------------------------
// Register Map (Register definition see "Fusion 10 Spec F10A-0102 v1.5.pdf" chapter 3.6.2)
#define REGMAPSIZE                      0x12                           // size of Register Map

// Data Information Register
#define DATA_INFO_REG                   0x00                            // register address
#define DATA_INFO_REG_SIZE              1                               // count of bytes

#define ucNUMBER_OF_FINGERS()            (ucRegisterMap[DATA_INFO_REG]&0x03) // access macros

// First Touch Point Information Registers
#define TP0_INFO_REG                    0x01                            // register address
#define TP0_INFO_REG_SIZE                6                               // count of bytes

#define uiTP0_X0_POSITION()                ((ucRegisterMap[TP0_INFO_REG+0]*256)+ucRegisterMap[TP0_INFO_REG+1]) // X0 Position
#define uiTP0_Y0_POSITION()                ((ucRegisterMap[TP0_INFO_REG+2]*256)+ucRegisterMap[TP0_INFO_REG+3]) // Y0 Position
#define ucTP0_PRESSURE_VALUE()            (ucRegisterMap[TP0_INFO_REG+4]) // First Touch Point Pressure Value
#define bTEST_TP0_TIP_SWITCH()            (0x01==(ucRegisterMap[TP0_INFO_REG+5]&0x0F)) // Tip Switch
#define ucTP0_TOUCH_ID()                ((ucRegisterMap[TP0_INFO_REG+5]&0xF0)>>4) // Touch ID

// Second Touch Point Information Registers
#define TP1_INFO_REG                    0x07                            // register address
#define TP1_INFO_REG_SIZE                6                               // count of bytes

#define uiTP1_X1_POSITION()                ((ucRegisterMap[TP1_INFO_REG+0]*256)+ucRegisterMap[TP1_INFO_REG+1]) // X0 Position
#define uiTP1_Y1_POSITION()                ((ucRegisterMap[TP1_INFO_REG+2]*256)+ucRegisterMap[TP1_INFO_REG+3]) // Y0 Position
#define ucTP1_PRESSURE_VALUE()            (ucRegisterMap[TP1_INFO_REG+4]) // First Touch Point Pressure Value
#define bTEST_TP1_TIP_SWITCH()            (0x01==(ucRegisterMap[TP1_INFO_REG+5]&0x0F)) // Tip Switch
#define ucTP1_TOUCH_ID()                ((ucRegisterMap[TP1_INFO_REG+5]&0xF0)>>4) // Touch ID

// Firmware Version
#define FIRMWARE_REG                    0x0E                            // register address //ok//
#define FIRMWARE_REG_SIZE                2                               // count of bytes  //ok//

#define uiFM_VREG                        ((ucRegisterMap[FIRMWARE_REG+0]*256)+ucRegisterMap[FIRMWARE_REG+1])  // Firmware Version as 2 Byte

// Reset Control Register
#define RESET_CTRL_REG                  0x10                            // register address
#define RESET_CTRL_REG_SIZE             1                               // count of bytes

#define vSTART_RESET()                    ucRegisterMap[RESET_CTRL_REG]=0x80 // start reset
#define RESET_WAIT_TIME                    130                                // wait time after start reset min 125ms

// Scan Complete Register
#define SCAN_COMPL_REG                   0x11                            // register address
#define SCAN_COMPL_REG_SIZE              1                               // count of bytes

#define vCLEAR_SCAN_COMPL()                (ucRegisterMap[SCAN_COMPL_REG]=0x00) // clear scan complete
#define bTEST_SCAN_COMPL()                (0x01==ucRegisterMap[SCAN_COMPL_REG]) // test scan complete

// Register sizes see "Fusion 10 Spec F10A-0102 v1.5.pdf" chapter 3.6.2
#define TOUCHREGBLK_SIZE                 DATA_INFO_REG_SIZE+TP0_INFO_REG_SIZE+TP1_INFO_REG_SIZE  // register blk for touchs


//==== privat type definitions ====================================================================
typedef struct
{
    UINT8    u8NuFinger;                                                    // Number of Fingers

    // first touch point information
    UINT16    u16Px1;                                                        // X Position
    UINT16    u16Py1;                                                        // Y Position
    BOOL    bTipSwitch1;                                                // Tip Switch
    UINT8    u8TouchID1;                                                    // Touch ID

    // second touch point information
    UINT16    u16Px2;                                                        // X Position
    UINT16    u16Py2;                                                        // Y Position
    BOOL    bTipSwitch2;                                                // Tip Switch
    UINT8    u8TouchID2;                                                    // Touch ID
}I2C_DEVICE;



//==== privat members =============================================================================
static UCHAR                             ucRegisterMap[REGMAPSIZE];        // copy of Register Map

static DWORD                            dwI2C_Speed;
static DWORD                            dwI2C_Adr;

static DWORD                            dwINT_Pin;
static DWORD                            dwINT_Sig_Inv;
static DWORD                            dwRESET_Pin;
static DWORD                            dwRESET_PostDly;
static DWORD                            dwRESET_Lin_Inv;
static DWORD                            dwINT_GPIONumber;
static DWORD                            dwRESET_GPIONumber;

static HANDLE                           hI2C_Event;                     // for "interrupts"
static DWORD                            dwI2C_SysIntr;
static DWORD                            dwHwAdaptPrio;                    // task priority

#ifdef TRIGGER_ON
static DWORD                            dwTRIGGER_Pin;
static DWORD                            dwTRIGGER_GPIONumber;
#endif

//------ Trace Buffers ----------------------------------------------------------------------------
#ifdef TRACE_ON

static struct                                                            // trace for last index
{
    CETCHINP_IOCTL                    TchInpCtrl;                            // touch events and counts
    DWORD                            dwTicks;
}TraceTchBuf[200];

static UINT iTchIndex;                                                    // index for trace buffer

#endif




//==== privat methods definitions ================================================================
/// @internal
/// Read out parts of "Work Mode Registers" out of the touch controller of Fusion10 or Fusion10 and place in the internal copy (array)
/// The index of this array is equal to the addresses of the "Work Mode Registers" in the touch controller
/// @param[in]    sStartRegister     The device's i2c address (0..127)
/// @param[in]    iNumber     Number of bytes to read
/// @retval        TRUE  success
/// @retval        FALSE error, a message to the console shows the reason
/// @caveats    none
static BOOL bReadInRegister(SHORT sStartRegister,UINT iNumber)
{
    BOOL bSuccess;
    LONG lErr;
    INT i;
    DWORD dwAlingBuffer[REGMAPSIZE/4+1];    // the Tegra I2C library needs an alingment of 4 for the buffer
                                            // therefore all I2C read or write use the alignment buffer and copy it to the not alignment place
    bSuccess = FALSE;

    if (FALSE==GetI2CLock(1000))
    {
        lErr = GetLastError();
        DEBUGMSG(ZONE_ERROR,(TEXT("MultiTchHwAdapt:bReadInRegister() illegale response in GetI2Lock() %d\r\n"),lErr));
        Sleep(50);
    }
    else
    {
        i = 0;
        while (1==1)                        // retry, sometimes on T30 there are some curious error
        {
            if (FALSE == I2CBurstRead((UCHAR)dwI2C_Adr,(UCHAR*)dwAlingBuffer,sStartRegister,iNumber))
            {
                if (2<i)
                {
                    lErr = GetLastError();
                    DEBUGMSG(ZONE_ERROR,(TEXT("MultiTchHwAdapt:bReadInRegister() illegale response in I2CBurstRead() %d\r\n"),lErr));
                    break;
                }
                i++;
            }
            else
            {
                bSuccess = TRUE;
                memcpy(&ucRegisterMap[sStartRegister],dwAlingBuffer,iNumber);    // copie the alligment buffer to orginal place
                break;
            }
        }
        ReleaseI2CLock();
    }
    return bSuccess;


}

/// @internal
/// Write out parts of internal register copy (array) to the "Work Mode Registers" of the touch controller Fusion10
/// The index of this array is equal to the adrresses of the "Work Mode Registers" in the touch controller
/// @param[in]    sStartRegister     The device's i2c address (0..127)
/// @param[in]    iNumber     Number of bytes to read
/// @retval        TRUE  success
/// @retval        FALSE error, a message to the console shows the reason
/// @caveats    none
static BOOL bWriteOutRegister(SHORT sStartRegister,UINT iNumber) //ok//
{
  BOOL bSuccess;
    LONG lErr;
    INT i;

    DWORD dwAlingBuffer[REGMAPSIZE/4+1];    // the Tegra I2C library needs an alingment of 4 for the buffer
                                            // therefore all I2C read or write use the alignment buffer and copy it to the not alignment place

    memcpy(dwAlingBuffer,&ucRegisterMap[sStartRegister],iNumber);    // copie the alligment buffer to orginal place

    bSuccess = FALSE;

    if (FALSE==GetI2CLock(1000))
    {
        lErr = GetLastError();
        DEBUGMSG(ZONE_ERROR,(TEXT("MultiTchHwAdapt:bWriteOutRegister() illegale response in GetI2Lock() %d\r\n"),lErr));
        Sleep(50);
    }
    else
    {
        i = 0;
        while (1==1)                        // retry, sometimes on T30 there are some curious error
        {
            if (FALSE == I2CBurstWrite((UCHAR)dwI2C_Adr,(UCHAR*)dwAlingBuffer,sStartRegister,iNumber))
            {
                if (2<i)
                {
                    lErr = GetLastError();
                    DEBUGMSG(ZONE_ERROR,(TEXT("MultiTchHwAdapt:bWriteOutRegister() illegale response in I2CBurstWrite() %d\r\n"),lErr));
                    break;
                }
                i++;
            }
            else
            {
                bSuccess = TRUE;
                break;
            }
        }
        ReleaseI2CLock();
    }

    return bSuccess;
}


/// Test the level of the INT line
/// @param[in]    none
/// @param[out]   none
/// @retval       TRUE  INT line activ
/// @retval       FALSE INT line passiv
/// @caveats      none
static BOOL bTstINTFusion(void) //ok//
{
    if (0==dwINT_Sig_Inv)                                            // interrupt input is not inverted
    {
        return (1==GetGPIOLevel(dwINT_GPIONumber));                    // interrupt input allready high means there are pending touchs
    }
    else
    {
        return (0==GetGPIOLevel(dwINT_GPIONumber));                    // interrupt input allready low means there are pending touchs
    }
}

/// Read out all the registers containing touch data of the touch controller
/// @param[in]    psI2CDat = pointer to the touch events data
/// @param[out]   psI2CDat = pointer to the touch events data (in case of error the data are set to no touches)
/// @retval       none
/// @caveats      none
static void vGetControllerData(I2C_DEVICE *psI2CDat)
{

    DEBUGMSG(ZONE_FUNCTION, ( TEXT("MultiTchHwAdapt:vGetControllerData()+\r\n" )) );

    if (FALSE==bReadInRegister(DATA_INFO_REG,TOUCHREGBLK_SIZE))
    {
        ERRORMSG(ZONE_ERROR, (TEXT("MultiTchHwAdapt:vGetControllerData() I2C read failed\r\n")));
        psI2CDat->u8NuFinger     = 0;                                    // clean all data
        psI2CDat->u16Px1         = 0;
        psI2CDat->u16Py1         = 0;
        psI2CDat->bTipSwitch1     = FALSE;
        psI2CDat->u8TouchID1     = 0;
        psI2CDat->u16Px2         = 0;
        psI2CDat->u16Py2         = 0;
        psI2CDat->bTipSwitch2     = FALSE;
        psI2CDat->u8TouchID2     = 0;
        goto out;
    }

    psI2CDat->u8NuFinger    = ucNUMBER_OF_FINGERS();                    // Number of Fingers

    psI2CDat->u16Px1        = uiTP0_Y0_POSITION();                        // originaly, x and y are changed (no idea why)
    psI2CDat->u16Py1        = uiTP0_X0_POSITION();

    psI2CDat->bTipSwitch1    = bTEST_TP0_TIP_SWITCH();                    // Tip Switch (TRUE = finger on touch)
    psI2CDat->u8TouchID1    = ucTP0_TOUCH_ID();                            // Touch ID

    psI2CDat->u16Px2        = uiTP1_Y1_POSITION();                        // originaly, x and y are changed (no idea why)
    psI2CDat->u16Py2        = uiTP1_X1_POSITION();

    psI2CDat->bTipSwitch2    = bTEST_TP1_TIP_SWITCH();                    // Tip Switch
    psI2CDat->u8TouchID2    = ucTP1_TOUCH_ID();                            // Touch ID

out:
    DEBUGMSG(ZONE_FUNCTION,(TEXT("MultiTchHwAdapt:vGetControllerData() -\r\n" )));

    return;
}




//==== public preprocessor definitions ===========================================================
//==== public members ============================================================================
//==== public methods definitions ================================================================

/// Set the configuration data according to the registry entry or to the default value
void vI2C_ReadConfiguration(void)
{
    HKEY    hKey;
    DWORD   dwSize;
    LONG    lError;

    // set default registry values
    dwI2C_Speed            = I2C_400KB;                                    // I2C speed
    dwI2C_Adr           = I2C_ADR;                                        // I2C address of the TangoC48 chip

    dwINT_Pin            = INT_PIN;                                        // pin for Int
    dwINT_Sig_Inv        = INT_SIG_INV;                                    // Int signal invertetd
    dwRESET_Pin            = RESET_PIN;                                    // pin for Reset
    dwRESET_PostDly        = RESET_POSTDLY;                                // delay after the "RESET Line" is released
    dwRESET_Lin_Inv        = RESET_LIN_INV;                                // Reset line inverted
    dwHwAdaptPrio        = HWADAPTPRIO;                                    // Task priority
    dwINT_GPIONumber    = 0;                                            // GPIO pin for Int
    dwRESET_GPIONumber  = 0;                                            // GPIO pin for Reset

#ifdef TRIGGER_ON
    dwTRIGGER_Pin    = dwTRIGGER_PIN;
#endif

    // lock if there are settings in the registry
    lError = RegOpenKeyEx(HKEY_LOCAL_MACHINE, L"Software\\Toradex\\MultiTchHwAdapt", 0, KEY_ALL_ACCESS, &hKey);
    if (lError == ERROR_SUCCESS)                                        // only if the registry key exitsts
    {
        dwSize=sizeof(DWORD);
        RegQueryValueEx(hKey, L"Int_Pin", NULL, NULL, (LPBYTE)&dwINT_Pin, &dwSize); // get the GPIONumber for interrupt
        RegQueryValueEx(hKey, L"Int_Signal_Inv", NULL, NULL, (LPBYTE)&dwINT_Sig_Inv, &dwSize); // get the GPIONumber for interrupt active level
        RegQueryValueEx(hKey, L"Reset_Pin", NULL, NULL, (LPBYTE)&dwRESET_Pin, &dwSize); // get the GPIONumber for reset
        RegQueryValueEx(hKey, L"Reset_Post_Delay", NULL, NULL, (LPBYTE)&dwRESET_PostDly, &dwSize); // get the pso delay for reset
        RegQueryValueEx(hKey, L"Reset_Line_Inv", NULL, NULL, (LPBYTE)&dwRESET_Lin_Inv, &dwSize); // get the inversion for reset
        RegQueryValueEx(hKey, L"I2CSpeed", NULL, NULL, (LPBYTE)&dwI2C_Speed,&dwSize); // get the I2C Speed
        RegQueryValueEx(hKey, L"I2CAddress", NULL, NULL, (LPBYTE)&dwI2C_Adr,&dwSize); // get the I2C Address
        RegQueryValueEx(hKey, L"HwAdaptPriority", NULL, NULL, (LPBYTE)&dwHwAdaptPrio,&dwSize); // get the priority
        RegCloseKey (hKey);
    }
}

/// Init the I2C communcation incl.the GIPO for the "INT" and "RST" pin of the touch controller
BOOL bI2C_Init(void) //ok//
{
    BOOL        bSuccess;
    DWORD       dwVerMaj;
    DWORD       dwVerMin;
    DWORD       dwVerBuild;
    DWORD       dwIRq;
    DWORD       dwEdge;

    PIN_INSTANCE sGPIO;

    //------ I2C init ----------------------------------------------------------------------------
    I2cIntGetLibVersion(&dwVerMaj, &dwVerMin, &dwVerBuild);

    DEBUGMSG(ZONE_INFO,(TEXT("MultiTchAppDriv:I2C Library Version=%d.%d Build=%d\r\n"),dwVerMaj, dwVerMin, dwVerBuild));

    bSuccess = FALSE;

    if (TRUE==I2CInit())
    {
        switch (dwI2C_Speed)
        {
            case I2C_100KB:
                bSuccess = GetI2CLock(500);
                if (TRUE==bSuccess)
                {
                    I2CSetSpeed(I2C_100KBPS);                            // set speed
                    ReleaseI2CLock();
                }
                break;

            case I2C_400KB:
                bSuccess = GetI2CLock(500);
                if (TRUE==bSuccess)
                {
                    I2CSetSpeed(I2C_400KBPS);                            // set speed
                    ReleaseI2CLock();
                }
                break;

            default:
                break;
        }
    }

    //------ GPIO init ----------------------------------------------------------------------------
    if (TRUE==bSuccess)
    {
        bSuccess = FALSE;

        if (TRUE == GetGPIOFromPin(dwINT_Pin, FALSE, &sGPIO))
        {
            if (0xFF != sGPIO.inst2)                                    // function on a second GPIO exist
            {
                SetGPIOAltFn((DWORD)sGPIO.inst2, -1, DIR_IN);            // set it to input
            }
            dwINT_GPIONumber = sGPIO.inst1;                                // use inst1 as GPIONumber

            if (TRUE == GetGPIOFromPin(dwRESET_Pin, FALSE, &sGPIO))
            {
                if (0xFF != sGPIO.inst2)                                // function on a second GPIO exist
                {
                    SetGPIOAltFn((DWORD)sGPIO.inst2, -1, DIR_IN);        // set it to input
                }
                dwRESET_GPIONumber = sGPIO.inst1;                        // use inst1 as GPIONumber
                bSuccess = TRUE;
            }
            else
            {
                DEBUGMSG(ZONE_ERROR,(TEXT("Multi-Touch HW-Adaption_Init:bI2C_Init() look up Reset_GPIONumber fails\r\n")));
            }
        }
        else
        {
            DEBUGMSG(ZONE_ERROR,(TEXT("Multi-Touch HW-Adaption_Init:bI2C_Init() look up Int_GPIONumber fails\r\n")));
        }
    }

    if (TRUE==bSuccess)
    {
        bSuccess = FALSE;

        SetGPIOAltFn(dwRESET_GPIONumber, -1, DIR_OUT);
        if (0==dwRESET_Lin_Inv)                                            // Reset not inverted
        {
            SetGPIOLevel (dwRESET_GPIONumber,0);                        // Reset activ / no power (vInitTangoC48())
        }
        else
        {
            SetGPIOLevel (dwRESET_GPIONumber,1);                        // Reset activ / no power (vInitTangoC48())
        }

        SetGPIOAltFn(dwINT_GPIONumber, -1, DIR_IN);                        // Set INT as Input

        dwIRq=GetGPIOIrq(dwINT_GPIONumber);
        if(dwIRq)
        {
            // Set edge to trigger (Rising, Falling or both)
            // dwEdge=GPIO_EDGE_RISING | GPIO_EDGE_FALLING;                // (rising & falling edge)
            // dwEdge=GPIO_EDGE_FALLING;

            dwEdge=GPIO_EDGE_RISING;

            if(SetGPIOIrqEdge(dwINT_GPIONumber,dwEdge))
            {
                // Create an event to wait on
                hI2C_Event = CreateEvent(NULL, FALSE, FALSE, NULL);
                if (hI2C_Event!= NULL)
                {
                    // Get the SYSINTR that corresponds to dwIrq (non GPIO Irqs corresond to the bit numbers of the PXA27x Interrupt resisters)
                    dwI2C_SysIntr=RequestSysInterrupt(dwIRq);

                    if (dwI2C_SysIntr)
                    {
                        if (InterruptInitializeCompat(dwI2C_SysIntr,hI2C_Event,NULL,0))
                        {
                            // set interrupt priority
                            CeSetThreadPriority(GetCurrentThread(), dwHwAdaptPrio);
                            bSuccess = TRUE;
                        }
                    }
                }
            }
        }
    }

    if (FALSE==bSuccess)
    {
        DEBUGMSG(ZONE_ERROR,(TEXT("Multi-Touch HW-Adaption_Init:bI2C_Init() had failed\r\n")));
    }

#ifdef TRIGGER_ON
     if (TRUE==bSuccess)
    {
        // for Trigger Output
        if (TRUE == GetGPIOFromPin(dwTRIGGER_Pin, FALSE, &sGPIO))
        {
            if (0xFF != sGPIO.inst2)                                // function on a second GPIO exist
            {
                SetGPIOAltFn((DWORD)sGPIO.inst2, -1, DIR_IN);        // set it to input
            }
            dwTRIGGER_GPIONumber = sGPIO.inst1;                        // use inst1 as GPIONumber
            bSuccess = TRUE;
        }
        else
        {
            DEBUGMSG(ZONE_ERROR,(TEXT("MultiTchAppDriv:bI2C_Init() look up TRIGGER_GPIONumber fails\r\n")));
        }
        SetGPIOAltFn(dwTRIGGER_GPIONumber, -1, DIR_OUT);
        SetGPIOLevel (dwTRIGGER_GPIONumber,0);                        //  set it to low for start
    }
#endif

    return bSuccess;
}


/// Start the touch controller (release the RST/PWR line) start the commincation and and setup the
BOOL bInitFusion(void)
{
    if (0==dwRESET_Lin_Inv)                                                // Reset not inverted
    {
        SetGPIOLevel (dwRESET_GPIONumber,1);                            // Reset passiv / Power on)
    }
    else
    {
        SetGPIOLevel (dwRESET_GPIONumber,0);                            // Reset passiv / Power on)
    }

    Sleep(dwRESET_PostDly);                                                // wait until the reset sequence of the touch controller is over

    //-------- touch controller --------------------------------------------------------------------
    if (FALSE==bReadInRegister(FIRMWARE_REG,FIRMWARE_REG_SIZE))
    {
        return FALSE;
    }
    else
    {
        vCLEAR_SCAN_COMPL();                                            // clear scan complete bit

        if (FALSE==bWriteOutRegister(SCAN_COMPL_REG,SCAN_COMPL_REG_SIZE))
        {
            return FALSE;
        }
        else
        {
            // Set the proper state for the next interrupt
            InterruptDoneCompat(dwI2C_SysIntr);
            return TRUE;
        }
    }
}


/// Release the current interrupt and wait for the next interrupt with timeout
BOOL bWaitTouchDataINT(void) //ok//
{
    BOOL bRet;

    if (WAIT_TIMEOUT == WaitForSingleObject(hI2C_Event,COMTST_INT))    // wait for input interrupt or timeout
    {
        bRet = FALSE;
    }
    else
    {
        bRet = TRUE;
    }

    //++    InterruptDoneCompat(dwI2C_SysIntr);                            // interrupt quit
    return bRet;
}

/// Test the communication by reading out the vedors's chp id
BOOL bCommunicationTestFusion(void) //ok//
{
    BOOL bRet;

    if (TRUE==bTstINTFusion())                                            // Interrupt activ in idle ?
    {
        bRet = FALSE;
    }
    else
    {
        if (FALSE==bReadInRegister(FIRMWARE_REG,FIRMWARE_REG_SIZE))
        {
            bRet = FALSE;
        }
        else
        {
            bRet = TRUE;
        }
    }

    // Set the proper state for the next interrupt.
    InterruptDoneCompat(dwI2C_SysIntr);                                    // interrupt quit
    return bRet;
}

/// Get the touch events from the controller
void vTouchPanelGetPoint(PTCHINPUT pInput,INT *pCnt)
{
    // static variables for previous evnts
    static BOOL bEventPrimary        = FALSE;
    static BOOL bPrevTipSwitchID1    = FALSE;
    static BOOL bPrevTipSwitchID2    = FALSE;
    static BOOL bTouchEventPrimary    = FALSE;
    static INT  s_xPrevious;
    static INT  s_yPrevious;

    I2C_DEVICE sI2CDat;                                                    // data structure of the controller data

    INT        iFiCnt;

    BOOL bPrevTipSwitch_tmp;

    PTCHINPUT pInput_tmp;                                                // only for debug
    pInput_tmp = pInput;


    DEBUGMSG(ZONE_FUNCTION||ZONE_INFO, (TEXT("MultiTchAppDriv:vTouchPanelGetPoint() +\r\n")));

    // reset pointer for TOUCHEVENTF_PRIMARY flag handling
    *pCnt = 0;

    // clear all data for the call back to the MDD
    memset(pInput, 0, MAXTOUCHES*sizeof(TCHINPUT));                        // clear array for call back function

    vGetControllerData(&sI2CDat);                                    // get the data from the touch Controller

    // Set the proper state for the next interrupt.
    InterruptDoneCompat(dwI2C_SysIntr);                                    // interrupt quit

    // write scan_complete to touch controller register to set the interrupt line passiv.
    vCLEAR_SCAN_COMPL();                                            // clear scan complete bit
    if (FALSE==bWriteOutRegister(SCAN_COMPL_REG,SCAN_COMPL_REG_SIZE))
    {
        DEBUGMSG(ZONE_ERROR, (TEXT("MultiTchAppDriv:vTouchPanelGetPoint() I2C write failed\r\n")));
        goto out;
    }

    iFiCnt = sI2CDat.u8NuFinger;                                        // get th ecount of fingers

    if ((0<iFiCnt)&&(3>iFiCnt))                                            // count of finges range vaid ?
    {
        //-------- generate the first data for the call back --------------------------------------------
        if (1==sI2CDat.u8TouchID1)                                        // choose the rigth previous state
            bPrevTipSwitch_tmp = bPrevTipSwitchID1;
        else
            bPrevTipSwitch_tmp = bPrevTipSwitchID2;

        if (sI2CDat.bTipSwitch1)                                        // Touch Controller reports finger on screen
        {
            pInput->x        = sI2CDat.u16Px1;
            pInput->y        = sI2CDat.u16Py1;
            s_xPrevious        = sI2CDat.u16Px1;
            s_yPrevious        = sI2CDat.u16Py1;

            pInput->dwID    = sI2CDat.u8TouchID1;

            if (bPrevTipSwitch_tmp)                                        // previous allready a finger on the screen
            {
                pInput->dwFlags    = TOUCHEVENTF_MOVE|TOUCHEVENTF_INRANGE;
            }
            else                                                        // previous as not no finger on the screen
            {
                pInput->dwFlags    = TOUCHEVENTF_DOWN|TOUCHEVENTF_INRANGE;
                bPrevTipSwitch_tmp    = TRUE;
            }

            *pCnt = 1;
            pInput++;                                                    // prepared for the next entry
        }
        else
        {
            if (bPrevTipSwitch_tmp)                                        // previous allready a finger on the screen
            {
                pInput->x        = s_xPrevious;
                pInput->y        = s_yPrevious;

                pInput->dwID    = sI2CDat.u8TouchID1;

                pInput->dwFlags    = TOUCHEVENTF_UP;
                bPrevTipSwitch_tmp    = FALSE;

                *pCnt = 1;
                pInput++;                                                // prepared for the next entry
            }
            // else no finger on screen and no tocuh
        }

        if (1==sI2CDat.u8TouchID1)                                        // save the previous state
            bPrevTipSwitchID1 = bPrevTipSwitch_tmp;
        else
            bPrevTipSwitchID2 = bPrevTipSwitch_tmp;


        //-------- generate the second data for the call back -------------------------------------

        if (1==sI2CDat.u8TouchID2)                                        // choose the rigth previous state
            bPrevTipSwitch_tmp = bPrevTipSwitchID1;
        else
            bPrevTipSwitch_tmp = bPrevTipSwitchID2;

        if (2==iFiCnt)                                                    // count of finges range vaid ?
        {
            if (sI2CDat.bTipSwitch2)                                    // Touch Controller reports finger on screen
            {
                pInput->x        = sI2CDat.u16Px2;
                pInput->y        = sI2CDat.u16Py2;
                s_xPrevious        = sI2CDat.u16Px2;
                s_yPrevious        = sI2CDat.u16Py2;

                pInput->dwID    = sI2CDat.u8TouchID2;

                if (bPrevTipSwitch_tmp)                                    // previous allready a finger on the screen
                {
                    pInput->dwFlags    = TOUCHEVENTF_MOVE|TOUCHEVENTF_INRANGE;
                }
                else                                                    // previous as not no finger on the screen
                {
                    pInput->dwFlags    = TOUCHEVENTF_DOWN|TOUCHEVENTF_INRANGE;
                    bPrevTipSwitch_tmp    = TRUE;
                }

                *pCnt = 2;
            }
            else
            {
                if (bPrevTipSwitch_tmp)                                    // previous allready a finger on the screen
                {
                    pInput->x        = s_xPrevious;
                    pInput->y        = s_yPrevious;

                    pInput->dwID    = sI2CDat.u8TouchID2;

                    pInput->dwFlags    = TOUCHEVENTF_UP;
                    bPrevTipSwitch_tmp    = FALSE;

                    *pCnt = 2;
                }
                // else no finger on screen and no tocuh
            }

        }

        if (1==sI2CDat.u8TouchID2)                                    // save the previous state
            bPrevTipSwitchID1 = bPrevTipSwitch_tmp;
        else
            bPrevTipSwitchID2 = bPrevTipSwitch_tmp;



#if 0
        //-------- auswertung to print ------------------------------------------------------------
        switch (sI2CDat.u8NuFinger)
        {
            case 0:
                DEBUGMSG(ZONE_INFO,(TEXT("MultiTchAppDriv:vTouchPanelGetPoint() Fusion10 no touches\r\n")));
                break;
            case 1:
                DEBUGMSG(ZONE_INFO,(TEXT("MultiTchAppDriv:vTouchPanelGetPoint() Fusion10 FirstTouch TipSwich(%d),TouchID(%d), X(%d), Y(%d)\r\n" ),sI2CDat.bTipSwitch1,sI2CDat.u8TouchID1,sI2CDat.u16Px1,sI2CDat.u16Py1));
                break;
            case 2:
                DEBUGMSG(ZONE_INFO,(TEXT("MultiTchAppDriv:vTouchPanelGetPoint() Fusion10 FirstTouch TipSwich(%d),TouchID(%d), X(%d), Y(%d)\r\n" ),sI2CDat.bTipSwitch1,sI2CDat.u8TouchID1,sI2CDat.u16Px1,sI2CDat.u16Py1));
                DEBUGMSG(ZONE_INFO,(TEXT("MultiTchAppDriv:vTouchPanelGetPoint() Fusion10 SecondTouch TipSwich(%d),TouchID(%d), X(%d), Y(%d)\r\n" ),sI2CDat.bTipSwitch2,sI2CDat.u8TouchID2,sI2CDat.u16Px2,sI2CDat.u16Py2));
                break;
            default:
                DEBUGMSG(ZONE_INFO,( TEXT("MultiTchAppDriv:vTouchPanelGetPoint() Fusion10 illegale  of finges (%d)"),sI2CDat.u8NuFinger));
                break;
        }

        switch (*pCnt)
        {
            case 1:
                DEBUGMSG(ZONE_INFO,(TEXT("MultiTchAppDriv:vTouchPanelGetPoint() Input[0] dwID(%d), dwFlags(%x), X(%d), Y(%d)\r\n" ),pInput_tmp->dwID, pInput_tmp->dwFlags, pInput_tmp->x, pInput_tmp->y));
                break;

            case 2:
                DEBUGMSG(ZONE_INFO,(TEXT("MultiTchAppDriv:vTouchPanelGetPoint() Input[0] dwID(%d), dwFlags(%x), X(%d), Y(%d)\r\n" ),pInput_tmp->dwID, pInput_tmp->dwFlags, pInput_tmp->x, pInput_tmp->y));
                pInput_tmp++;
                DEBUGMSG(ZONE_INFO,(TEXT("MultiTchAppDriv:vTouchPanelGetPoint() Input[1] dwID(%d), dwFlags(%x), X(%d), Y(%d)\r\n" ),pInput_tmp->dwID, pInput_tmp->dwFlags, pInput_tmp->x, pInput_tmp->y));
                break;

            default:
                DEBUGMSG(ZONE_INFO,(TEXT("MultiTchAppDriv:vTouchPanelGetPoint() illegale pCint (%d)\r\n" ),*pCnt));
        }
#endif

    }
    else
    {
        DEBUGMSG(ZONE_INFO,(TEXT("MultiTchAppDriv:vTouchPanelGetPoint() Fusion10 no valid count of finges (%d)\r\n" ),iFiCnt));
    }

out:
    DEBUGMSG(ZONE_FUNCTION||ZONE_INFO,(TEXT("MultiTchAppDriv:vTouchPanelGetPoint() -\r\n")));
    return;
}

/// Reset on or power off the touch controller
void vDeInitFusion(void)                                                // stop (reset/powerdown) the Touch Controller
{
    if (0==dwRESET_Lin_Inv)                                                // Reset not inverted
    {
        SetGPIOLevel (dwRESET_GPIONumber,0);                            // Reset activ or no power
    }
    else
    {
        SetGPIOLevel (dwRESET_GPIONumber,1);                            // Reset activ or no power
    }

    ReleaseI2CLock();
}


/// Stop the I2C communication (release interrrupt, mutex, stop the I2C driver etc)
void vI2C_DeInit(void)
{
    // GPIO disabel
    InterruptDoneCompat(dwI2C_SysIntr);
    InterruptDisableCompat(dwI2C_SysIntr);
    ReleaseSysIntr(dwI2C_SysIntr);

    // I2C driver
    ReleaseI2CLock();
    I2CDeInit();
}

//------ Trace Functions --------------------------------------------------------------------------
#ifdef TRACE_ON

// reset index and whole touch buffer
void TraceTchClear(void)
{

    // erase trace buffer
    iTchIndex = 1000;
    memset(&TraceTchBuf[0], 0, sizeof(TraceTchBuf));
}

// write touch data to the controller buffer
void TraceTchPut(CETCHINP_IOCTL *pTchBuf)
{
    if (1000 == iTchIndex)
        iTchIndex=0;

    if (iTchIndex < 200)
    {
        TraceTchBuf[iTchIndex].TchInpCtrl    = *pTchBuf;
        TraceTchBuf[iTchIndex].dwTicks        = GetTickCount();
        iTchIndex++;
    }
}

// print out the whole touch data buffer
void TraceTchPrint(void)
{
    UINT i,j,cnt;

    if (1000 == iTchIndex)
        return;

    for (i = 0; i <iTchIndex; i++)
    {
        cnt = TraceTchBuf[i].TchInpCtrl.iCnt;

        DEBUGMSG(ZONE_WARN,(TEXT("TraceTchIndex=%d, Touch Count=%d, Time=%d\r\n" ),i,cnt,TraceTchBuf[i].dwTicks));
        for (j = 0; j <cnt; j++)
        {
            DEBUGMSG(ZONE_WARN,(TEXT("TraceTchIndex=%d, ID=%d, Flags=%d, x=%d, y=%d\r\n" ),i,TraceTchBuf[i].TchInpCtrl.TchInp[j].dwID,TraceTchBuf[i].TchInpCtrl.TchInp[j].dwFlags,TraceTchBuf[i].TchInpCtrl.TchInp[j].x,TraceTchBuf[i].TchInpCtrl.TchInp[j].y));
            Sleep(10);
        }
    }
    DEBUGMSG(ZONE_WARN,(TEXT("\r\n" )));
}
#endif

#ifdef TRIGGER_ON
/// Set the trigger output to high
void vSetTriggerHigh(void)
{
    SetGPIOLevel (dwTRIGGER_GPIONumber,1);
}

/// Set the trigger output to low
void vSetTriggerLow(void)
{
    SetGPIOLevel (dwTRIGGER_GPIONumber,0);
}
#endif










